"""
Module to run tests on scripts
"""
import os
import sys
import glob
import shutil
from IPython.terminal.embed import embed

from configobj import ConfigObj

import pytest

import numpy as np

import matplotlib
matplotlib.use('agg')  # For Travis

from pypeit.scripts.parse_calib_id import ParseCalibID
from pypeit.scripts.setup import Setup
from pypeit.scripts.run_pypeit import RunPypeIt
from pypeit.tests.tstutils import dev_suite_required
from pypeit import specobjs


def data_path(filename):
    data_dir = os.path.join(os.path.dirname(__file__), 'files')
    return os.path.join(data_dir, filename)

@dev_suite_required
def test_run_pypeit_calib_only():
    # Get the directories
    rawdir = os.path.join(os.environ['PYPEIT_DEV'], 'RAW_DATA', 'shane_kast_blue', '600_4310_d55')
    assert os.path.isdir(rawdir), 'Incorrect raw directory'

    # File list
    all_files = {
        'arcs': ['b1.fits.gz'],
        'flats': ['b11.fits.gz', 'b12.fits.gz', 'b13.fits.gz'],
        'bias': ['b21.fits.gz', 'b22.fits.gz', 'b23.fits.gz'],
    }
    all_masters = ['MasterArc_A_1_01.fits', 'MasterTiltimg_A_1_01.fits',
                   'MasterBias_A_1_01.fits',
                   'MasterTilts_A_1_01.fits', 'MasterEdges_A_1_01.fits.gz',
                   'MasterFlat_A_1_01.fits',
                   'MasterWaveCalib_A_1_01.fits']

    # Just get a few files
    for ss, sub_files, masters in zip(range(3),
            [['arcs', 'flats', 'bias'], ['arcs', 'bias'], ['flats', 'bias']],
            [all_masters, ['MasterArc_A_1_01.fits', 'MasterTiltimg_A_1_01.fits'],
             ['MasterEdges_A_1_01.fits.gz']]):
        # Grab the subset
        files = []
        for sub_file in sub_files:
            files += all_files[sub_file]
        #
        testrawdir = os.path.join(rawdir, 'TEST')
        if os.path.isdir(testrawdir):
            shutil.rmtree(testrawdir)
        os.makedirs(testrawdir)
        for f in files:
            shutil.copy(os.path.join(rawdir, f), os.path.join(testrawdir, f))

        outdir = os.path.join(os.getenv('PYPEIT_DEV'), 'REDUX_OUT_TEST')

        # For previously failed tests
        if os.path.isdir(outdir):
            shutil.rmtree(outdir)

        # Run the setup
        sargs = Setup.parse_args(['-r', testrawdir, '-s', 'shane_kast_blue', '-c all', '-o',
                                  '--output_path', outdir])
        Setup.main(sargs)

        # Change to the configuration directory and set the pypeit file
        configdir = os.path.join(outdir, 'shane_kast_blue_A')
        pyp_file = os.path.join(configdir, 'shane_kast_blue_A.pypeit')
        assert os.path.isfile(pyp_file), 'PypeIt file not written.'

        # Perform the calib-only reduction
        pargs = RunPypeIt.parse_args([pyp_file, '-c', '-r', configdir])
        RunPypeIt.main(pargs)

        # Test!
        for master_file in masters:
            assert os.path.isfile(os.path.join(configdir, 'Masters', master_file)
                                  ), 'Master File {:s} missing!'.format(master_file)

        # Now test parse_calib_id
        if ss == 0:
            pargs2 = ParseCalibID.parse_args([pyp_file])
            calib_dict = ParseCalibID.main(pargs2)
            assert isinstance(calib_dict, dict)
            assert len(calib_dict) > 0
            assert calib_dict['1']['A_1_01']['arc']['raw_files'][0] == 'b1.fits.gz'

        # Clean-up
        shutil.rmtree(outdir)
        shutil.rmtree(testrawdir)

def test_run_pypeit():

    # Just get a few files
    testrawdir = data_path('')

    #outdir = os.path.join(os.getenv('PYPEIT_DEV'), 'REDUX_OUT_TEST')
    outdir = data_path('REDUX_OUT_TEST')

    # For previously failed tests
    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    # Run the setup
    sargs = Setup.parse_args(['-r', testrawdir+'b', '-s', 
                              'shane_kast_blue', '-c all', '-o', 
                              '--output_path', outdir])
    Setup.main(sargs)

    # Change to the configuration directory and set the pypeit file
    configdir = os.path.join(outdir, 'shane_kast_blue_A')
    pyp_file = os.path.join(configdir, 'shane_kast_blue_A.pypeit')
    assert os.path.isfile(pyp_file), 'PypeIt file not written.'

    # Try to run with -m and -o
    pargs = RunPypeIt.parse_args([pyp_file, '-o', '-m', '-r', configdir])
    RunPypeIt.main(pargs)

    # #########################################################
    # Test!!
    # Files exist
    assert os.path.isfile(os.path.join(configdir, 'Science', 'spec2d_b27-J1217p3905_KASTb_20150520T045733.560.fits'))

    # spec1d
    spec1d_file = os.path.join(configdir, 'Science', 'spec1d_b27-J1217p3905_KASTb_20150520T045733.560.fits')
    assert os.path.isfile(spec1d_file)
    specObjs = specobjs.SpecObjs.from_fitsfile(spec1d_file)
    
    # Check RMS
    assert specObjs[0].WAVE_RMS < 0.02  # difference must be less than 0.02 pixels

    # Flexure
    assert abs(-0.03 - specObjs[0].FLEX_SHIFT_TOTAL) < 0.1  # difference must be less than 0.1 pixels

    # Helio
    assert abs(specObjs[0].VEL_CORR - 0.9999261685542624) < 1.0E-10

    # Now re-use those master files
    pargs = RunPypeIt.parse_args([pyp_file, '-o', '-r', configdir])
    RunPypeIt.main(pargs)

    # Clean-up
    shutil.rmtree(outdir)





"""
Module for Bok/B&C specific methods.

.. include:: ../include/links.rst
"""
from pkg_resources import resource_filename

import numpy as np

from astropy.time import Time

from pypeit import msgs
from pypeit import telescopes
from pypeit import io
from pypeit.core import framematch
from pypeit.spectrographs import spectrograph
from pypeit.core import parse
from pypeit.images import detector_container


class BokBCSpectrograph(spectrograph.Spectrograph):
    """
    Child to handle BOK specific code
    """
    ndet = 1
    telescope = telescopes.BokTelescopePar()
    name = 'bok_bc'
    camera = 'BC'
    comment = 'Bok B&C spectrometer'
    header_name = 'Bok B&C spectrometer'
    supported = True

    def configuration_keys(self):
        """
        Return the metadata keys that define a unique instrument
        configuration.

        This list is used by :class:`~pypeit.metadata.PypeItMetaData` to
        identify the unique configurations among the list of frames read
        for a given reduction.

        Returns:
            :obj:`list`: List of keywords of data pulled from file headers
            and used to constuct the :class:`~pypeit.metadata.PypeItMetaData`
            object.
        """
        return ['dispname', 'decker', 'binning', 'dispangle']

    def init_meta(self):
        """
        Define how metadata are derived from the spectrograph files.

        That is, this associates the ``PypeIt``-specific metadata keywords
        with the instrument-specific header cards using :attr:`meta`.
        """
        self.meta = {}
        # Required (core)
        self.meta['ra'] = dict(ext=0, card='RA')
        self.meta['dec'] = dict(ext=0, card='DEC')
        self.meta['target'] = dict(ext=0, card='OBJECT')
        self.meta['decker'] = dict(ext=0, card='APERTURE')
        self.meta['binning'] = dict(card=None, compound=True)
        self.meta['mjd'] = dict(card=None, compound=True)
        self.meta['exptime'] = dict(ext=0, card='EXPTIME')
        self.meta['airmass'] = dict(ext=0, card='AIRMASS')
        # Extras for config and frametyping
        self.meta['dispname'] = dict(ext=0, card='DISPERSE')
        self.meta['dispangle'] = dict(ext=0, card='TILTPOS', rtol=1e-3)
        self.meta['idname'] = dict(ext=0, card='OBJECT')
        self.meta['instrument'] = dict(ext=0, card='INSTRUME')
        # used for arc and continuum lamps
        self.meta['lampstat01'] = dict(ext=0, card=None, compound=True)

    def compound_meta(self, headarr, meta_key):
        """
        Methods to generate metadata requiring interpretation of the header
        data, instead of simply reading the value of a header card.

        Args:
            headarr (:obj:`list`):
                List of `astropy.io.fits.Header`_ objects.
            meta_key (:obj:`str`):
                Metadata keyword to construct.

        Returns:
            object: Metadata value read from the header(s).
        """
        if meta_key == 'binning':
            binspatial = headarr[0]['CCDBIN1']
            binspec = headarr[0]['CCDBIN2']
            return parse.binning2string(binspatial, binspec)
            #return parse.binning2string(binspec, binspatial)
        elif meta_key == 'mjd':
            """
            Need to combine 'DATE-OBS' and 'UT' headers and then use astropy to make an mjd.
            """
            date = headarr[0]['DATE-OBS']
            ut = headarr[0]['UT']
            ttime = Time(f"{date}T{ut}", format='isot')
            return ttime.mjd
        elif meta_key == 'lampstat01':
            """
            If the comparison mirror is in, there will be a 'COMPLAMP' header entry containing the lamps
            that are turned on. However, if the comparison mirror is out, then this header entry doesn't exist.
            So need to test for it and set to 'Off' if it's not there.
            """
            if 'COMPLAMP' in headarr[0]:
                return headarr[0]['COMPLAMP']
            else:
                return 'off'
        msgs.error("Not ready for this compound meta")

    def pypeit_file_keys(self):
        """
        Define the list of keys to be output into a standard ``PypeIt`` file.

        Returns:
            :obj:`list`: The list of keywords in the relevant
            :class:`~pypeit.metadata.PypeItMetaData` instance to print to the
            :ref:`pypeit_file`.
        """
        return super().pypeit_file_keys() + ['slitwid']


    def get_detector_par(self, det, hdu=None):
        """
        Return metadata for the selected detector.

        Args:
            det (:obj:`int`):
                1-indexed detector number.
            hdu (`astropy.io.fits.HDUList`_, optional):
                The open fits file with the raw image of interest.  If not
                provided, frame-dependent parameters are set to a default.

        Returns:
            :class:`~pypeit.images.detector_container.DetectorContainer`:
            Object with the detector metadata.
        """
        # Binning
        # TODO: Could this be detector dependent??
        binning = '1,1' if hdu is None \
                    else self.get_meta_value(self.get_headarr(hdu), 'binning')

        # Detector 1
        detector_dict = dict(
            binning         = binning,
            det             = 1,
            dataext         = 0,
            specaxis        = 1,
            specflip        = False,
            spatflip        = False,
            #platescale      = 15.0/18.0,
            platescale      = 0.2,
            darkcurr        = 5.4,
            saturation      = 65535.,
            nonlinear       = 1.0,
            mincounts       = -1e10,
            numamplifiers   = 1,
            gain            = np.atleast_1d(1.5),
            ronoise         = np.atleast_1d(3.0),
            datasec         = np.atleast_1d('[:,1:1200]')
            #datasec         = np.atleast_1d('[250:650,1:1200]'),
            )
        return detector_container.DetectorContainer(**detector_dict)

    @classmethod
    def default_pypeit_par(cls):
        """
        Return the default parameters to use for this instrument.
        
        Returns:
            :class:`~pypeit.par.pypeitpar.PypeItPar`: Parameters required by
            all of ``PypeIt`` methods.
        """
        par = super().default_pypeit_par()

        # Turn off illumflat
        turn_off = dict(use_illumflat=False, use_biasimage=False, use_overscan=False,
                        use_darkimage=False)
        par.reset_all_processimages_par(**turn_off)
        # TODO: Note this comment doesn't match up with what's actually done...
        # Require dark images to be subtracted from the flat images used for
        # tracing, pixelflats, and illumflats
        par['calibrations']['traceframe']['process']['use_darkimage'] = False
        par['calibrations']['pixelflatframe']['process']['use_darkimage'] = False
        par['calibrations']['illumflatframe']['process']['use_darkimage'] = False

        # Ignore PCA
        par['calibrations']['slitedges']['sync_predict'] = 'nearest'

        # JFH Is this correct?
        # Processing steps
        #turn_off = dict(use_overscan=False)
        #par.reset_all_processimages_par(**turn_off)

        # Turn off the overscan
        #for ftype in par['calibrations'].keys():
        #    try:
        #        par['calibrations'][ftype]['process']['overscan'] = 'none'
        #    except (TypeError, KeyError):
        #        pass
        par['scienceframe']['process']['use_overscan'] = False
        # Make a bad pixel mask
        par['calibrations']['bpm_usebias'] = False
        # Set pixel flat combination method
        par['calibrations']['pixelflatframe']['process']['combine'] = 'median'
        # Change the wavelength calibration method
        par['calibrations']['wavelengths']['method'] = 'holy-grail'
        #par['calibrations']['wavelengths']['method'] = 'reidentify'
        par['calibrations']['wavelengths']['lamps'] = ['NeI', 'ArI', 'ArII', 'HeI']
        # Wavelengths
        # 1D wavelength solution
        par['calibrations']['wavelengths']['rms_threshold'] = 0.5
        par['calibrations']['wavelengths']['sigdetect'] = 5.
        par['calibrations']['wavelengths']['fwhm']= 5.0

        #par['calibrations']['wavelengths']['n_first'] = 3
        #par['calibrations']['wavelengths']['n_final'] = 5
        #par['calibrations']['wavelengths']['sigdetect'] = 10.0
        #par['calibrations']['wavelengths']['wv_cen'] = 4859.0
        #par['calibrations']['wavelengths']['disp'] = 0.2
        # Do not flux calibrate
        par['fluxcalib'] = None
        # Set the default exposure time ranges for the frame typing
        par['calibrations']['biasframe']['exprng'] = [None, 1]
        par['calibrations']['darkframe']['exprng'] = [999999, None]     # No dark frames
        par['calibrations']['pinholeframe']['exprng'] = [999999, None]  # No pinhole frames
        par['calibrations']['arcframe']['exprng'] = [None, 120]
        par['calibrations']['standardframe']['exprng'] = [None, 120]
        par['scienceframe']['exprng'] = [90, None]

        # Extraction
        par['reduce']['skysub']['bspline_spacing'] = 0.8
        par['reduce']['skysub']['no_poly'] = True
        par['reduce']['skysub']['bspline_spacing'] = 0.6
        par['reduce']['skysub']['joint_fit'] = False
        par['reduce']['skysub']['global_sky_std']  = False

        par['reduce']['extraction']['sn_gauss'] = 4.0
        par['reduce']['findobj']['sig_thresh'] = 5.0
        par['reduce']['skysub']['sky_sigrej'] = 5.0
        par['reduce']['findobj']['find_trim_edge'] = [5,5]

        # cosmic ray rejection parameters for science frames
        par['scienceframe']['process']['sigclip'] = 5.0
        par['scienceframe']['process']['objlim'] = 2.0

        # Sensitivity function parameters
        par['sensfunc']['polyorder'] = 7

        # Do not correct for flexure
        par['flexure']['spec_method'] = 'skip'

        return par

    def bpm(self, filename, det, shape=None, msbias=None):
        """
        Generate a default bad-pixel mask.

        Even though they are both optional, either the precise shape for
        the image (``shape``) or an example file that can be read to get
        the shape (``filename`` using :func:`get_image_shape`) *must* be
        provided.

        Args:
            filename (:obj:`str` or None):
                An example file to use to get the image shape.
            det (:obj:`int`):
                1-indexed detector number to use when getting the image
                shape from the example file.
            shape (tuple, optional):
                Processed image shape
                Required if filename is None
                Ignored if filename is not None
            msbias (`numpy.ndarray`_, optional):
                Master bias frame used to identify bad pixels

        Returns:
            `numpy.ndarray`_: An integer array with a masked value set
            to 1 and an unmasked value set to 0.  All values are set to
            0.
        """

        # Call the base-class method to generate the empty bpm
        bpm_img = super().bpm(filename, det, shape=shape, msbias=msbias)

        if det == 1:
            msgs.info("Using hard-coded BPM for Bok B&C")

            bpm_img[:, -1] = 1

        else:
            msgs.error(f"Invalid detector number, {det}, for Bok B&C (only one detector).")

        return bpm_img

    def config_specific_par(self, scifile, inp_par=None):
        """
        Modify the ``PypeIt`` parameters to hard-wired values used for
        specific instrument configurations.

        Args:
            scifile (:obj:`str`):
                File to use when determining the configuration and how
                to adjust the input parameters.
            inp_par (:class:`~pypeit.par.parset.ParSet`, optional):
                Parameter set used for the full run of PypeIt.  If None,
                use :func:`default_pypeit_par`.

        Returns:
            :class:`~pypeit.par.parset.ParSet`: The PypeIt parameter set
            adjusted for configuration specific parameter values.
        """
        # Start with instrument wide
        par = super().config_specific_par(scifile, inp_par=inp_par)

        # Wavelength calibrations
        #if self.get_meta_value(scifile, 'dispname') == 'R1200B':
        #    par['calibrations']['wavelengths']['reid_arxiv'] = 'wht_isis_blue_1200_4800.fits'

        return par

    def check_frame_type(self, ftype, fitstbl, exprng=None):
        """
        Check for frames of the provided type.

        Args:
            ftype (:obj:`str`):
                Type of frame to check. Must be a valid frame type; see
                frame-type :ref:`frame_type_defs`.
            fitstbl (`astropy.table.Table`_):
                The table with the metadata for one or more frames to check.
            exprng (:obj:`list`, optional):
                Range in the allowed exposure time for a frame of type
                ``ftype``. See
                :func:`pypeit.core.framematch.check_frame_exptime`.

        Returns:
            `numpy.ndarray`_: Boolean array with the flags selecting the
            exposures in ``fitstbl`` that are ``ftype`` type frames.
        """
        good_exp = framematch.check_frame_exptime(fitstbl['exptime'], exprng)
        if ftype in ['science']:
            science = []
            for obj in fitstbl['idname'].tolist():
                science.append(not (("Dome Flat" in obj) or ("STANDARD" in obj) or ("HIP05" in obj) or ("HZ" in obj) or ("G191" in obj) or ("PG0220" in obj)))
            return good_exp & (fitstbl['lampstat01'] == 'off') & science
        if ftype in ['standard']:
            standard = []
            for obj in fitstbl['idname'].tolist():
                standard.append(("STANDARD" in obj) or ("HIP05" in obj) or ("HZ" in obj) or ("G191" in obj) or ("PG0220" in obj) )
            return good_exp & (fitstbl['lampstat01'] == 'off') & (fitstbl['idname'] != 'Dome Flat') & standard
        if ftype == 'bias':
            bias = []
            for obj in fitstbl['idname'].tolist():
                bias.append(("BIAS" in obj) or ("Bias" in obj))
            return good_exp & (fitstbl['lampstat01'] == 'off') & bias
        if ftype in ['pixelflat', 'trace']:
            flat = []
            for obj in fitstbl['idname'].tolist():
                flat.append(("Dome Flat" in obj))
            return good_exp & (fitstbl['lampstat01'] == 'off') & flat
        if ftype in ['pinhole', 'dark']:
            # Don't type pinhole or dark frames
            return np.zeros(len(fitstbl), dtype=bool)
        if ftype in ['arc', 'tilt']:
            return good_exp & (fitstbl['lampstat01'] != 'off')
        msgs.warn('Cannot determine if frames are of type {0}.'.format(ftype))
        return np.zeros(len(fitstbl), dtype=bool)






#
# See top-level LICENSE file for Copyright information
#
# -*- coding: utf-8 -*-
"""
This module contains code for collating multiple 1d spectra source object.
"""

import copy
import os.path

import numpy as np
from astropy.time import Time
import astropy.units as u
from astropy.coordinates import SkyCoord, Angle

from pypeit import specobjs
from pypeit.spectrographs.util import load_spectrograph
from pypeit import msgs

class SourceObject:

    """A group of reduced spectra from the same source object. This contains
    the information needed to coadd the spectra and archive the metadata.

    An instance is initiated with the first spectra of the group. Additional
    spectra can be compared with this object to see if it matches using the
    match method, and are added to it if they do.

    Args:
        spec1d_obj (:obj:`pypeit.specobj.SpecObj`):
            The initial spectra of the group as a SpecObj.
        spec1d_header (:obj:`astropy.io.fits.Header`): 
            The header for the first spec1d file in the group.
        spec1d_file (str): Filename of the first spec1d file in the group.
        spectrograph (:obj:`pypeit.spectrographs.spectrograph.Spectrograph`): 
            The spectrograph that was used to take the data.
        match_type (str): How spectra should be compared. 'ra/dec' means the
            spectra should be compared using the sky coordinates in RA and DEC.
            'pixel' means the spectra should be compared by the spatial pixel
            coordinates in the image.
        path (str): The path to place a coadded file. Defaults to the current directory.

    Attributes:
        spec_obj_list (list of :obj:`pypeit.spectrographs.spectrograph.Spectrograph`):
            The list of spectra in the group as SpecObj objects.
        spec1d_file_list (list of str): 
            The pathnames of the spec1d files in the group.
        spec1d_header_list: (list of :obj:`astropy.io.fits.Header`):
            The headers of the spec1d files in the group
    """
    def __init__(self, spec1d_obj, spec1d_header, spec1d_file, spectrograph, match_type, path=""):
        self.spec_obj_list = [spec1d_obj]
        self.spec1d_file_list = [spec1d_file]
        self.spec1d_header_list = [spec1d_header]
        self._spectrograph = spectrograph
        self.match_type = match_type

        if (match_type == 'ra/dec'):
            try:
                self.coord = SkyCoord(spec1d_obj.RA, spec1d_obj.DEC, unit='deg')
            except Exception as e:
                msgs.error(f"Cannot do ra/dec matching on {spec1d_obj.NAME}, could not read RA/DEC.")
        else:
            self.coord = spec1d_obj['SPAT_PIXPOS']

        self.coaddfile = os.path.join(path, self.build_coadd_file_name())

    @classmethod
    def build_source_objects(cls, spec1d_files, match_type, path=""):
        """Build a list of SourceObjects from a list of spec1d files. There will be one SourceObject per
        SpecObj in the resulting list (i.e. no combining or collating is done by this method).

        Args:
            spec1d_files (list of str): List of spec1d filenames
            match_type (str):           What type of matching the SourceObjects will be configured for.
                                        Must be either 'ra/dec' or 'pixel'
            path (str):                 The path where the coadded file should be placed. Defaults to 
                                        the current directory.

        Returns: 
            list of :obj:`SourceObject`: A list of uncollated SourceObjects with one SpecObj per SourceObject.
        """
        result = []
        for spec1d_file in spec1d_files:            
            sobjs = specobjs.SpecObjs.from_fitsfile(spec1d_file)
            spectrograph = load_spectrograph(sobjs.header['PYP_SPEC'])
            for sobj in sobjs:
                result.append(SourceObject(sobj, sobjs.header, spec1d_file, spectrograph, match_type, path))
    
        return result

    def build_coadd_file_name(self):
        """Build the output file name for coadding.
        The filename convention is J<hmsdms+dms>_<instrument name>_<YYYYMMDD>.fits
        when matching by RA/DEC and SPAT_<spatial position>_<instrument name>_<YYYYMMDD>.fits
        when matching by pixel position..

        Currently instrument_name is taken from spectrograph.camera

        Returns: 
            str:  The name of the coadd output file.
        """
        time_portion = Time(self.spec1d_header_list[0]['MJD'], format="mjd").strftime('%Y%m%d')
        if self.match_type == 'ra/dec':
            coord_portion = 'J' + self.coord.to_string('hmsdms', sep='', precision=2).replace(' ', '')
        else:
            coord_portion = self.spec_obj_list[0]['NAME'].split('_')[0]
        instrument_name = self._spectrograph.camera

        return f'{coord_portion}_{instrument_name}_{time_portion}.fits'

    def _config_key_match(self, header):
        """
        Check to see if the configuration keys from a spec1d file match the
        ones for this SourceObject.

        Args:
            header (:obj:`astropy.io.fits.Header`):
                Header from a spec1d file.

        Returns:
            bool: True if the configuration keys match, 
                  false if they do not.
        """
        # Make sure the spectrograph matches
        if 'PYP_SPEC' not in header or header['PYP_SPEC'] != self._spectrograph.name:
            return False

        first_header = self.spec1d_header_list[0]

        for key in self._spectrograph.configuration_keys():
            # Ignore "decker" because it's valid to coadd spectra with different slit masks
            if key != "decker":
                if key not in first_header and key not in header:
                    # Both are missing the key, this is ok
                    continue
                elif key not in first_header or key not in header:
                    # One has a value and the other doesn't so they don't match
                    return False

                if first_header[key] != header[key]:
                    return False

        # No mismatches were found
        return True

    def match(self, spec_obj, spec1d_header, tolerance, unit = u.arcsec):
        """Determine if a SpecObj matches this group within the given tolerance.
        This will also compare the configuration keys to make sure the SpecObj
        is compatible with the ones in this SourceObject.

        Args:
            spec_obj (:obj:`pypeit.specobj.SpecObj`): 
                The SpecObj to compare with this SourceObject.

            spec1d_header (:obj:`astropy.io.fits.Header`):
                The header from the spec1d that dontains the SpecObj.
                
            tolerance (float): 
                Maximum distance that two spectra can be from each other to be 
                considered to be from the same source. Measured in floating
                point pixels or as an angular distance (see ``unit1`` argument).

            unit (:obj:`astropy.units.Unit`):
                Units of ``tolerance`` argument if match_type is 'ra/dec'. 
                Defaults to arcseconds. Igored if match_type is 'pixel'.

        Returns:
            bool: True if the SpecObj matches this group,
                  False otherwise.
        """

        if not self._config_key_match(spec1d_header):
            return False

        if self.match_type == 'ra/dec':
            coord2 = SkyCoord(ra=spec_obj.RA, dec=spec_obj.DEC, unit='deg')
            return self.coord.separation(coord2) <= Angle(tolerance, unit=unit)
        else:
            coord2 =spec_obj['SPAT_PIXPOS'] 
            return np.fabs(coord2 - self.coord) <= tolerance

    def combine(self, other_source_object):
        """Combine this SourceObject with another. The two objects must be from the
        same spectrograph and use the same match type.

        Args:
        other_source_object (:obj:`SourceObject`): The other object to combine with.

        Returns:
            (:obj:`SourceObject`): This SourceObject, now combined with other_source_object.
        """
        
        if other_source_object._spectrograph.name != self._spectrograph.name or \
           other_source_object.match_type != self.match_type:
           msgs.error(f"Can't append incompatible source objects. {self.spectrograph.name}/{self.match_type} does not match {other_source_object.spectrograph.name}/{other_source_object.match_type}")

        self.spec_obj_list += other_source_object.spec_obj_list
        self.spec1d_file_list += other_source_object.spec1d_file_list
        self.spec1d_header_list += other_source_object.spec1d_header_list

        return self

def collate_spectra_by_source(source_list, tolerance, unit=u.arcsec):
    """Given a list of spec1d files from PypeIt, group the spectra within the
    files by their source object. The grouping is done by comparing the 
    position of each spectra (using either pixel or RA/DEC) using a given tolerance.

    Args:
        source_list (list of :obj:`SourceObject`): A list of source objects, one
            SpecObj per object, ready for collation.
        tolerance (float): 
            Maximum distance that two spectra can be from each other to be 
            considered to be from the same source. Measured in floating
            point pixels or as an angular distance (see ``unit`` argument).
        unit (:obj:`astropy.units.Unit`):
            Units of ``tolerance`` argument if match_type is 'ra/dec'. 
            Defaults to arcseconds. Ignored if match_type is 'pixel'.

    Returns:
        (list of `obj`:SourceObject): The collated spectra as SourceObjects.

    """

    collated_list = []
    for source in source_list:

        # Search for a collated SourceObject that matches this one.
        # If one can't be found, treat this as a new collated SourceObject.
        found = False
        for collated_source in collated_list:
            if  collated_source.match(source.spec_obj_list[0], 
                                      source.spec1d_header_list[0],
                                      tolerance, unit):

                collated_source.combine(source)
                found = True

        if not found:
            collated_list.append(copy.deepcopy(source))

    return collated_list



"""
Module to run tests on simple fitting routines for arrays
"""
import os

import pytest

import numpy as np

from linetools.spectra.io import readspec

import pypeit
from pypeit.core import flexure, arc

from pypeit.spectrographs.util import load_spectrograph
from pypeit.tests.tstutils import cooked_required

def data_path(filename):
    data_dir = os.path.join(os.path.dirname(__file__), 'files')
    return os.path.join(data_dir, filename)

from IPython import embed

#dev_path = os.getenv('PYPEIT_DEV')
#lris_path =  os.path.join(dev_path, 'REDUX_OUT/Keck_LRIS_blue/multi_600_4000_d560')
#trc_file = os.path.join(lris_path, 'Masters', 'MasterTrace_A_1_01.fits')
#tilts_file = os.path.join(lris_path, 'Masters', 'MasterTilts_A_1_01.fits')
#spec2dfile = os.path.join(lris_path, 'Science', 'spec2d_b170320_2083-c17_60L._LRISb_2017Mar20T055336.211.fits')

#def test_flex_lris():
#    dpath = '/scratch/REDUX/Keck/LRIS/2020jan28/keck_lris_red_C'
#    spec2dfile = os.path.join(dpath, 'Science', 'spec2d_LR.20200128.57449-frb1907B_LRISr_2020Jan28T155729.779.fits')
#    tilts_file = os.path.join(dpath, 'Masters', 'MasterTilts_C_1_01.fits')
#    slits_file = os.path.join(dpath, 'Masters', 'MasterSlits_C_1_01.fits.gz')
#    # Load
#    hdul = fits.open(spec2dfile)
#    sciimg = hdul[1].data
#    slits = slittrace.SlitTraceSet.from_file(slits_file)
#    tilts = wavetilts.WaveTilts.from_file(tilts_file)
#
#    flexure.spat_flexure_shift(sciimg, slits)


def test_flex_shift():
    # Dummy slf
    # Read spectra
    obj_spec = readspec(data_path('obj_lrisb_600_sky.fits'))
    arx_file = pypeit.__path__[0]+'/data/sky_spec/sky_LRISb_600.fits'
    arx_spec = readspec(arx_file)
    arx_lines = arc.detect_lines(arx_spec.flux.value)

    # Call
    flex_dict = flexure.spec_flex_shift(obj_spec, arx_spec, arx_lines, mxshft=60)

#    # Apply
#    from scipy import interpolate
#    print(flex_dict['shift'])
#    npix = len(obj_spec.wavelength)
#    x = np.linspace(0., 1., npix)
#    f = interpolate.interp1d(x, obj_spec.wavelength.value, bounds_error=False,
#                             fill_value="extrapolate")
#    new_wave = f(x+flex_dict['shift']/(npix-1))
#
#    from matplotlib import pyplot
#    pyplot.plot(arx_spec.wavelength, arx_spec.flux)
#    pyplot.plot(obj_spec.wavelength, obj_spec.flux)
#    pyplot.plot(new_wave, obj_spec.flux)
#    pyplot.show()
    assert np.abs(flex_dict['shift'] - 43.7) < 0.1

@cooked_required
def test_flex_multi():
   
    spec1d_file = os.path.join(os.getenv('PYPEIT_DEV'), 'Cooked', 'Science',
                            'spec1d_DE.20100913.22358-CFHQS1_DEIMOS_20100913T061231.334.fits')

    msFlex = flexure.MultiSlitFlexure(s1dfile=spec1d_file) 
    # Parameters
    keck_deimos = load_spectrograph('keck_deimos')
    par = keck_deimos.default_pypeit_par()
    msFlex.init(keck_deimos, par['flexure'])
    # Init                    
    outfile = data_path('tst_multi_flex.fits')
    # INITIAL SKY LINE STUFF
    msFlex.measure_sky_lines()
    # FIT SURFACES
    msFlex.fit_mask_surfaces()
    # Apply
    msFlex.update_fit()
    # QA
    #mask = header['TARGET'].strip()
    #fnames = header['FILENAME'].split('.')
    #root = mask+'_'+fnames[2]
    #mdFlex.qa_plots('./', root)

    # Write
    msFlex.to_file(outfile, overwrite=True)

    # Read
    msFlex2 = flexure.MultiSlitFlexure.from_file(outfile)
    msFlex2.to_file(outfile, overwrite=True)

    # Check
    assert np.all(np.isclose(msFlex2.fit_b, msFlex.fit_b))

    # Clean up
    if os.path.isfile(outfile):
        os.remove(outfile)


"""
Module for MMT/Blue Channel specific methods.

.. include:: ../include/links.rst
"""
import glob
import numpy as np
from astropy.io import fits
from astropy.time import Time

from pypeit import msgs
from pypeit import telescopes
from pypeit.core import framematch
from pypeit.par import pypeitpar
from pypeit.spectrographs import spectrograph
from pypeit.core import parse
from pypeit.images import detector_container

from pkg_resources import resource_filename

class MMTBlueChannelSpectrograph(spectrograph.Spectrograph):
    """
    Child to handle MMT/Blue Channel specific code
    """
    ndet = 1
    name = 'mmt_bluechannel'
    header_name = 'mmtbluechan'
    telescope = telescopes.MMTTelescopePar()
    camera = 'Blue_Channel'
    supported = True

    def get_detector_par(self, det, hdu=None):
        """
        Return metadata for the selected detector.

        .. warning::

            Many of the necessary detector parameters are read from the file
            header, meaning the ``hdu`` argument is effectively **required** for
            MMT/BlueChannel.  The optional use of ``hdu`` is only viable for
            automatically generated documentation.

        Args:
            det (:obj:`int`):
                1-indexed detector number.
            hdu (`astropy.io.fits.HDUList`_, optional):
                The open fits file with the raw image of interest.  If not
                provided, frame-dependent parameters are set to a default.

        Returns:
            :class:`~pypeit.images.detector_container.DetectorContainer`:
            Object with the detector metadata.
        """
        if hdu is None:
            binning = '1,1'
            gain = None
            ronoise = None
            darkcurr = None
            datasec = None
            oscansec = None
        else:
            binning = self.get_meta_value(self.get_headarr(hdu), 'binning')
            gain = np.atleast_1d(hdu[0].header['GAIN'])
            ronoise = np.atleast_1d(hdu[0].header['RDNOISE'])
            darkcurr = hdu[0].header['DARKCUR']
            datasec = np.atleast_1d(hdu[0].header['DATASEC'])
            oscansec = np.atleast_1d(hdu[0].header['BIASSEC'])

        # Detector 1
        detector_dict = dict(
            binning         = binning,
            det             = 1,
            dataext         = 0,
            specaxis        = 0,
            specflip        = False,
            spatflip        = False,
            xgap            = 0.,
            ygap            = 0.,
            ysize           = 1.,
            platescale      = 0.3,
            darkcurr        = darkcurr, #header['DARKCUR'],
            saturation      = 65535.,
            nonlinear       = 0.95,  # need to look up and update
            mincounts       = -1e10,
            numamplifiers   = 1,
            gain            = gain, #np.atleast_1d(header['GAIN']),
            ronoise         = ronoise, #np.atleast_1d(header['RDNOISE']),
            # note that the header entries use the binned sizes
# TODO: These aren't needed because the read_rawimage sets these directly,
# right?
            datasec         = datasec, #np.atleast_1d(header['DATASEC']),
            oscansec        = oscansec #np.atleast_1d(header['BIASSEC'])
        )

        return detector_container.DetectorContainer(**detector_dict)

    def init_meta(self):
        """
        Define how metadata are derived from the spectrograph files.

        That is, this associates the ``PypeIt``-specific metadata keywords
        with the instrument-specific header cards using :attr:`meta`.
        """
        self.meta = {}
        # Required (core)
        self.meta['ra'] = dict(ext=0, card='RA')
        self.meta['dec'] = dict(ext=0, card='DEC')
        self.meta['target'] = dict(ext=0, card='OBJECT')
        self.meta['decker'] = dict(ext=0, card='APERTURE')
        self.meta['dichroic'] = dict(ext=0, card='INSFILTE')
        self.meta['binning'] = dict(ext=0, card=None, compound=True)
        self.meta['mjd'] = dict(ext=0, card=None, compound=True)
        self.meta['exptime'] = dict(ext=0, card='EXPTIME')
        self.meta['airmass'] = dict(ext=0, card='AIRMASS')

        # Extras for config and frametyping
        self.meta['dispname'] = dict(ext=0, card='DISPERSE')
        self.meta['idname'] = dict(ext=0, card='IMAGETYP')

        # used for arc and continuum lamps
        self.meta['lampstat01'] = dict(ext=0, card=None, compound=True)
        self.meta['instrument'] = dict(ext=0, card='INSTRUME')

    def compound_meta(self, headarr, meta_key):
        """
        Methods to generate metadata requiring interpretation of the header
        data, instead of simply reading the value of a header card.

        Args:
            headarr (:obj:`list`):
                List of `astropy.io.fits.Header`_ objects.
            meta_key (:obj:`str`):
                Metadata keyword to construct.

        Returns:
            object: Metadata value read from the header(s).
        """
        if meta_key == 'binning':
            """
            Binning in blue channel headers is space-separated rather than comma-separated.
            """
            binspec, binspatial = headarr[0]['CCDSUM'].split()
            binning = parse.binning2string(binspec, binspatial)
            return binning
        elif meta_key == 'mjd':
            """
            Need to combine 'DATE-OBS' and 'UT' headers and then use astropy to make an mjd.
            """
            date = headarr[0]['DATE-OBS']
            ut = headarr[0]['UT']
            ttime = Time(f"{date}T{ut}", format='isot')
            return ttime.mjd
        elif meta_key == 'lampstat01':
            """
            If the comparison mirror is in, there will be a 'COMPLAMP' header entry containing the lamps
            that are turned on. However, if the comparison mirror is out, then this header entry doesn't exist.
            So need to test for it and set to 'Off' if it's not there.
            """
            if 'COMPLAMP' in headarr[0]:
                return headarr[0]['COMPLAMP']
            else:
                return 'off'
        
        msgs.error(f"Not ready for compound meta, {meta_key}, for MMT Blue Channel.")

    @classmethod
    def default_pypeit_par(cls):
        """
        Return the default parameters to use for this instrument.
        
        Returns:
            :class:`~pypeit.par.pypeitpar.PypeItPar`: Parameters required by
            all of ``PypeIt`` methods.
        """
        par = super().default_pypeit_par()

        # Wavelengths
        # 1D wavelength solution
        par['calibrations']['wavelengths']['rms_threshold'] = 0.5
        par['calibrations']['wavelengths']['sigdetect'] = 5.
        par['calibrations']['wavelengths']['fwhm']= 5.0
        # HeNeAr is by far most commonly used, though ThAr is used for some situations.
        par['calibrations']['wavelengths']['lamps'] = ['ArI', 'ArII', 'HeI', 'NeI']
        par['calibrations']['wavelengths']['method'] = 'holy-grail'

        # Processing steps
        turn_off = dict(use_biasimage=False, use_darkimage=False)
        par.reset_all_processimages_par(**turn_off)

        # Extraction
        par['reduce']['skysub']['bspline_spacing'] = 0.8
        par['reduce']['extraction']['sn_gauss'] = 4.0
        ## Do not perform global sky subtraction for standard stars
        par['reduce']['skysub']['global_sky_std']  = False

        # cosmic ray rejection parameters for science frames
        par['scienceframe']['process']['sigclip'] = 5.0
        par['scienceframe']['process']['objlim'] = 2.0

        # Set the default exposure time ranges for the frame typing

        # Appropriate exposure times for Blue Channel can vary a lot depending
        # on grating and wavelength. E.g. 300 and 500 line gratings need very
        # short exposures for flats to avoid saturation, but the 1200 and 832
        # can use much longer exposures due to the higher resolution and the
        # continuum lamp not being very bright in the blue/near-UV.
        par['calibrations']['pixelflatframe']['exprng'] = [None, 100]
        par['calibrations']['traceframe']['exprng'] = [None, 100]
        par['calibrations']['standardframe']['exprng'] = [None, 600]
        par['calibrations']['arcframe']['exprng'] = [10, None]
        par['calibrations']['darkframe']['exprng'] = [300, None]

        # less than 30 sec implies conditions are bright enough for scattered
        # light to be significant which affects the illumination of the slit.
        par['calibrations']['illumflatframe']['exprng'] = [30, None]

        # Need to specify this for long-slit data
        par['calibrations']['slitedges']['sync_predict'] = 'nearest'
        par['calibrations']['slitedges']['bound_detector'] = True

        # Sensitivity function parameters
        par['sensfunc']['polyorder'] = 7

        return par

    def bpm(self, filename, det, shape=None, msbias=None):
        """
        Generate a default bad-pixel mask.

        Even though they are both optional, either the precise shape for
        the image (``shape``) or an example file that can be read to get
        the shape (``filename`` using :func:`get_image_shape`) *must* be
        provided.

        Args:
            filename (:obj:`str` or None):
                An example file to use to get the image shape.
            det (:obj:`int`):
                1-indexed detector number to use when getting the image
                shape from the example file.
            shape (tuple, optional):
                Processed image shape
                Required if filename is None
                Ignored if filename is not None
            msbias (`numpy.ndarray`_, optional):
                Master bias frame used to identify bad pixels

        Returns:
            `numpy.ndarray`_: An integer array with a masked value set
            to 1 and an unmasked value set to 0.  All values are set to
            0.
        """
        # Call the base-class method to generate the empty bpm
        bpm_img = super().bpm(filename, det, shape=shape, msbias=msbias)

        if det == 1:
            msgs.info("Using hard-coded BPM for  Blue Channel")

            bpm_img[-1, :] = 1

        else:
            msgs.error(f"Invalid detector number, {det}, for MMT Blue Channel (only one detector).")

        return bpm_img

    def configuration_keys(self):
        """
        Return the metadata keys that define a unique instrument
        configuration.

        This list is used by :class:`~pypeit.metadata.PypeItMetaData` to
        identify the unique configurations among the list of frames read
        for a given reduction.

        Returns:
            :obj:`list`: List of keywords of data pulled from file headers
            and used to constuct the :class:`~pypeit.metadata.PypeItMetaData`
            object.
        """
        return ['dispname']

    def check_frame_type(self, ftype, fitstbl, exprng=None):
        """
        Check for frames of the provided type.

        Args:
            ftype (:obj:`str`):
                Type of frame to check. Must be a valid frame type; see
                frame-type :ref:`frame_type_defs`.
            fitstbl (`astropy.table.Table`_):
                The table with the metadata for one or more frames to check.
            exprng (:obj:`list`, optional):
                Range in the allowed exposure time for a frame of type
                ``ftype``. See
                :func:`pypeit.core.framematch.check_frame_exptime`.

        Returns:
            `numpy.ndarray`_: Boolean array with the flags selecting the
            exposures in ``fitstbl`` that are ``ftype`` type frames.
        """
        good_exp = framematch.check_frame_exptime(fitstbl['exptime'], exprng)
        if ftype == 'bias':
            return fitstbl['idname'] == 'zero'
        if ftype in ['science', 'standard']:
            return good_exp & (fitstbl['lampstat01'] == 'off') & (fitstbl['idname'] == 'object')
        if ftype in ['arc', 'tilt']:
            # should flesh this out to include all valid arc lamp combos
            return good_exp & (fitstbl['lampstat01'] != 'off') & (fitstbl['idname'] == 'comp') & (fitstbl['decker'] != '5.0x180')
        if ftype in ['trace', 'pixelflat']:
            # i think the bright lamp, BC, is the only one ever used for this. imagetyp should always be set to flat.
            return good_exp & (fitstbl['lampstat01'] == 'BC') & (fitstbl['idname'] == 'flat')
        if ftype in ['illumflat']:
            # these can be set to flat or object depending if they're twilight or dark sky
            return good_exp & (fitstbl['idname'] in ['flat', 'object']) & (fitstbl['lampstat01'] == 'off')

        msgs.warn('Cannot determine if frames are of type {0}.'.format(ftype))
        return np.zeros(len(fitstbl), dtype=bool)

    def get_rawimage(self, raw_file, det):
        """
        Read raw images and generate a few other bits and pieces
        that are key for image processing.

        Parameters
        ----------
        raw_file : :obj:`str`
            File to read
        det : :obj:`int`
            1-indexed detector to read

        Returns
        -------
        detector_par : :class:`pypeit.images.detector_container.DetectorContainer`
            Detector metadata parameters.
        raw_img : `numpy.ndarray`_
            Raw image for this detector.
        hdu : `astropy.io.fits.HDUList`_
            Opened fits file
        exptime : :obj:`float`
            Exposure time read from the file header
        rawdatasec_img : `numpy.ndarray`_
            Data (Science) section of the detector as provided by setting the
            (1-indexed) number of the amplifier used to read each detector
            pixel. Pixels unassociated with any amplifier are set to 0.
        oscansec_img : `numpy.ndarray`_
            Overscan section of the detector as provided by setting the
            (1-indexed) number of the amplifier used to read each detector
            pixel. Pixels unassociated with any amplifier are set to 0.
        """
        # Check for file; allow for extra .gz, etc. suffix
        fil = glob.glob(raw_file + '*')
        if len(fil) != 1:
            msgs.error("Found {:d} files matching {:s}".format(len(fil)))

        # Read FITS image
        msgs.info("Reading MMT Blue Channel file: {:s}".format(fil[0]))
        hdu = fits.open(fil[0])
        hdr = hdu[0].header

        # we're flipping FITS x/y to pypeit y/x here. pypeit wants blue on the
        # bottom, slit bottom on the right...
        rawdata = np.fliplr(hdu[0].data.astype(float).transpose())

        exptime = hdr['EXPTIME']

        # TODO Store these parameters in the DetectorPar.
        # Number of amplifiers
        detector_par = self.get_detector_par(det if det is not None else 1, hdu=hdu)
        numamp = detector_par['numamplifiers']

        # First read over the header info to determine the size of the output array...
        datasec = hdr['DATASEC']
        xdata1, xdata2, ydata1, ydata2 = np.array(parse.load_sections(datasec, fmt_iraf=False)).flatten()

        # Get the overscan section
        biassec = hdr['BIASSEC']
        xbias1, xbias2, ybias1, ybias2 = np.array(parse.load_sections(biassec, fmt_iraf=False)).flatten()

        # allocate output arrays and fill in with mask values
        rawdatasec_img = np.zeros_like(rawdata, dtype=int)
        oscansec_img = np.zeros_like(rawdata, dtype=int)

        # trim bad sections at beginning of data and bias sections
        rawdatasec_img[xdata1+2:xdata2, ydata1:ydata2-1] = 1
        oscansec_img[xbias1+2:xbias2, ybias1:ybias2-1] = 1

        return detector_par, rawdata, hdu, exptime, rawdatasec_img, oscansec_img





""" Generate the wavelength templates for P200/DBSP"""
import os

from pypeit.core.wavecal import templates

##
# DBSP red arm
##

# ##############################
def p200_dbsp_red_1200_7100_d68(overwrite=False):  # DBSPr 1200/7100 D68

    binspec = 1
    outroot = 'p200_dbsp_red_1200_7100_d68.fits'
    #
    ifiles = [0, 1]
    slits = [222, 0]  # Be careful with the order..
    lcut = [7740.]
    wfile0 = os.path.join(templates.template_path, 'P200_DBSP',
        'R1200_7100_D68', 'MasterWaveCalib_A_1_01_7600.fits')
    wfile1 = os.path.join(templates.template_path, 'P200_DBSP',
        'R1200_7100_D68', 'wvcalib_8200.fits')

    #
    templates.build_template([wfile0, wfile1], slits, lcut, binspec, outroot,
        lowredux=False, ifiles=ifiles, normalize=True, overwrite=overwrite)

# ##############################
def p200_dbsp_red_1200_9400_d55(overwrite=False):  # DBSPr 1200/9400 D55

    binspec = 1
    outroot = 'p200_dbsp_1200_9400_d55.fits'
    #
    ifiles = [0]
    slits = [0]  # Be careful with the order..
    lcut = None
    wfile0 = os.path.join(templates.template_path, 'P200_DBSP',
        'R1200_9400_D55', 'wvcalib_8800.fits')

    #
    templates.build_template([wfile0], slits, lcut, binspec, outroot,
        lowredux=False, ifiles=ifiles, normalize=True, overwrite=overwrite)

# ##############################
def p200_dbsp_red_600_10000_d55(overwrite=False):  # DBSPr 600/10000 D55

    binspec = 1
    outroot = 'p200_dbsp_red_600_10000_d55.fits'
    #
    ifiles = [0]
    slits = [221]  # Be careful with the order..
    lcut = None
    wfile0 = os.path.join(templates.template_path, 'P200_DBSP',
        'R600_10000_D55', 'MasterWaveCalib_A_1_01.fits')

    #
    templates.build_template([wfile0], slits, lcut, binspec, outroot,
        lowredux=False, ifiles=ifiles, normalize=True, overwrite=overwrite)

# ##############################
def p200_dbsp_red_316_7500_d55(overwrite=False):  # DBSPr 316/7500 D55

    binspec = 1
    outroot = 'p200_dbsp_red_316_7500_d55.fits'
    #
    ifiles = [0]
    slits = [221]  # Be careful with the order..
    lcut = None
    wfile0 = os.path.join(templates.template_path, 'P200_DBSP',
        'R316_7500_D55', 'P200_DBSP_Red.json')

    #
    templates.build_template([wfile0], slits, lcut, binspec, outroot,
        lowredux=False, ifiles=ifiles, normalize=True, overwrite=overwrite)

##
# DBSP blue arm
##

# ##############################
def p200_dbsp_blue_300_3990_d55(overwrite=False):  # DBSPb 300/3990 D55

    binspec = 1
    outroot = 'p200_dbsp_blue_300_3990_d55.fits'
    #
    ifiles = [0]
    slits = [0]  # Be careful with the order..
    lcut = None
    wfile0 = os.path.join(templates.template_path, 'P200_DBSP',
        'B300_3990_D55', 'wvcalib.fits')

    #
    templates.build_template([wfile0], slits, lcut, binspec, outroot,
        lowredux=False, ifiles=ifiles, normalize=True, overwrite=overwrite)

# ##############################
def p200_dbsp_blue_600_4000_d55(overwrite=False):  # DBSPb 600/4000 D55

    binspec = 1
    outroot = 'p200_dbsp_blue_600_4000_d55.fits'
    #
    ifiles = [0]
    slits = [231]  # Be careful with the order..
    lcut = None
    wfile0 = os.path.join(templates.template_path, 'P200_DBSP',
        'B600_4000_D55', 'P200_DBSP_Blue.json')

    #
    templates.build_template([wfile0], slits, lcut, binspec, outroot,
        lowredux=False, ifiles=ifiles, normalize=True, overwrite=overwrite)

# ##############################
def p200_dbsp_blue_1200_5000_d55(overwrite=False):  # DBSPb 1200/5000 D55

    binspec = 1
    outroot = 'p200_dbsp_blue_1200_5000_d55_4700.fits'
    #
    ifiles = [0]
    slits = [0]  # Be careful with the order..
    lcut = None
    wfile0 = os.path.join(templates.template_path, 'P200_DBSP',
        'B1200_5000_D55', 'wvcalib_4700.fits')

    #
    templates.build_template([wfile0], slits, lcut, binspec, outroot,
        lowredux=False, ifiles=ifiles, normalize=True, overwrite=overwrite)

# ##############################
def p200_dbsp_blue_1200_5000_d68(overwrite=False):  # DBSPb 1200/5000 D68

    binspec = 1
    outroot = 'p200_dbsp_blue_1200_5000_d68_6000.fits'
    #
    ifiles = [0]
    slits = [180]  # Be careful with the order..
    lcut = None
    wfile0 = os.path.join(templates.template_path, 'P200_DBSP',
        'B1200_5000_D68', 'MasterWaveCalib_A_1_01_6000.fits')

    #
    templates.build_template([wfile0], slits, lcut, binspec, outroot,
        lowredux=False, ifiles=ifiles, normalize=True, overwrite=overwrite)

if __name__ == '__main__':
    # p200_dbsp_red_316_7500_d55(overwrite=True)
    # p200_dbsp_red_600_10000_d55(overwrite=True)
    p200_dbsp_red_1200_7100_d68(overwrite=True)
